/*
** Copyright (c) 2002 D. Richard Hipp
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** General Public License for more details.
** 
** You should have received a copy of the GNU General Public
** License along with this library; if not, write to the
** Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA  02111-1307, USA.
**
** Author contact information:
**   drh@hwaci.com
**   http://www.hwaci.com/drh/
**
*******************************************************************************
**
** This file contains code used to generate convert plain text into
** similarly formated HTML.
*/
#include "config.h"
#include "format.h"
#include <time.h>


/*
** Return the number digits at the beginning of the string z.
*/
static int ndigit(const char *z){
  int i = 0;
  while( isdigit(*z) ){ i++; z++; }
  return i;
}

/*
** Scan white space at the beginning of z.  If there are two or
** more newline characters, then return the total number of whitespace
** characters in the prefix.  If the prefix contains fewer than two
** newlines, return 0.
*/
static int nspace(const char *z){
  int nNewline = 0;
  int count = 0;
  while( isspace(*z) ){
    if( *z=='\n' ){ nNewline++; }
    z++;
    count++;
  }
  return nNewline>=2 ? count : 0;
}

/*
** Check to see if *z points to the beginning of a Wiki page name.
** If it does, return the number of characters in that name.  If not,
** return 0.
**
** A Wiki page name contains only alphabetic characters.  The first
** letter must be capital and there must be at least one other capital
** letter in the word.  And every capital leter must be followed by
** one or more lower-case letters.
*/
int is_wiki_name(const char *z){
  int i;
  int nCap = 0;
  if( !isupper(z[0]) ) return 0;
  for(i=0; z[i]; i++){
    if( isupper(z[i]) ){
      if( !islower(z[i+1]) ) return 0;
      nCap++;
    }else if( !islower(z[i]) ){
      break;
    }
  }
  return nCap>=2 ? i : 0;
}

/*
** If *z points to a word or phrase that should be bold, return
** the number of characters in the word or phrase.  Otherwise return 0.
**
** A word should be bold if it begins and ends with and underscore.
*/
static int is_bold_word(const char *z){
  int i, c;
  if( z[0]!='_' || isspace(z[1]) ) return 0;
  for(i=1; z[i] && !isspace(z[i]); i++){}
  if( i>1 && ((c = z[i-1])=='.' || c==')' || c==':') ){ i--; }
  return z[i-1]=='_' ? i : 0;
}

/*
** If *z points to text of the form:  "N: " where N is a digit,
** then *z points to the start of an enumeration item.  Return 1.
** If *z does not point to such text, return 0.
*/
int is_enum(const char *z){
  if( !isdigit(z[0]) ) return 0;
  if( z[1]!=':' ) return 0;
  if( z[2]!=' ' && z[2]!='\t' ) return 0;
  return 1;
}

/*
** If *z points to text of the form:  "*: " 
** then *z points to the start of a bullet.  Return 1.
** If *z does not point to such text, return 0.
*/
int is_bullet(const char *z){
  if( z[0]!='*' ) return 0;
  if( z[1]!=':' ) return 0;
  if( z[2]!=' ' && z[2]!='\t' ) return 0;
  return 1;
}

/*
** Return the number of characters in the URL that begins
** at *z.
**
** Algorithm: Advance to the first whitespace character or until
** then end of the string.  Then back up over the following
** characters:  .)]}?!"':;,
*/
static int is_url(const char *z){
  int i;
  for(i=0; z[i] && !isspace(z[i]); i++){}
  while( i>0 ){
    switch( z[i-1] ){
      case '.':
      case ')':
      case ']':
      case '}':
      case '?':
      case '!':
      case '"':
      case '\'':
      case ':':
      case ';':
      case ',':
        i--;
        break;
      default:
        return i;
    }
  }
  return 0;
}

/*
** Output Wiki text while inserting the proper HTML control codes.
** The following formatting conventions are implemented:
**
**    *    Characters with special meaning to HTML are escaped.
**
**    *    Blank lines results in a paragraph break.
**
**    *    Paragraphs where the first line is indented by two or more
**         spaces are shown verbatim.  None of the following rules apply
**         to verbatim text.
**
**    *    Lines beginning with "*: " begin a bullet in a bullet list.
**
**    *    Lines beginning with "1: " begin an item in an enumerated list.
**
**    *    Words that begin and end with underscore (ex: _this_is_bold_)
**         are shown in bold.  Mid-word underscores are converted to
**         spaces.
**
**    *    Wiki pages names (Words in initial caps) are enclosed in an
**         appropriate hyperlink.
**
**    *    Words that begin with "http:" are enclosed in an appropriate
**         hyperlink.
**
**    *    Text of the form "#NNN" where NNN is a valid ticket number
**         is converted into a hyperlink to the corresponding ticket.
**
**    *    Text of the form "[NNN]" where NNN is a valid check-in number
**         becomes a hyperlink to the checkin.
*/
void output_wiki(const char *zText, const char *zLinkSuffix){
  int i, j, k;
  int inPRE = 0;
  int inUL = 0;
  int inOL = 0;
  int v;
  int wordStart = 1;
  static int once = 1;
  static int nTicket, nCommit;
  if( once ){
    nTicket = atoi(db_short_query("SELECT max(tn) FROM ticket"));
    nCommit = atoi(db_short_query("SELECT max(cn) FROM chng"));
    once = 0;
  }

  i = 0;
  while( isspace(zText[i]) ){ i++; }
  if( i>2 && zText[i-1]==' ' && zText[i-2]==' '){
    cgi_printf("<pre>");
    inPRE = 1;
  }else if( i==0 || zText[i-1]=='\n' ){
    if( zText[i]=='*' && zText[i+1]==':' && isspace(zText[i+2]) ){
      cgi_printf("<ul compact><li>");
      inUL = 1;
      zText += i+3;
      i = 0;
    }else if( isdigit(zText[i]) && zText[i+1]==':' && isspace(zText[i+2]) ){
      cgi_printf("<ol compact><li>");
      inOL = 1;
      zText += i+3;
      i = 0;
    }
  }
  while( zText[i] ){
    char *z;
    int c = zText[i];
    if( wordStart && !inPRE ){
      if( isupper(c) && (j = is_wiki_name(&zText[i]))>0 ){
        z = htmlize(zText, i);
        cgi_printf("%s", z);
        free(z);
        cgi_printf("<a href=\"wiki?p=%.*s%s\">%.*s</a>",
            j, &zText[i], zLinkSuffix, j, &zText[i]);
        zText += i+j;
        i = 0;
        wordStart = 0;
        continue;
      }
      if( c=='_' && (j = is_bold_word(&zText[i]))>0 ){
        z = htmlize(zText, i);
        cgi_printf("%s", z);
        free(z);
        z = htmlize(&zText[i+1], j-2);
        for(k=0; k<j-2; k++){
          if( z[k]=='_' ) z[k] = ' ';
        }
        cgi_printf("<b>%s</b>", z);
        free(z);
        zText += i+j;
        i = 0;
        wordStart = 0;
        continue;
      }
      if( c=='h' && (strncmp(&zText[i],"http:",5)==0 
                     || strncmp(&zText[i],"https:",6)==0)
             && (j=is_url(&zText[i]))>0 ){
        z = htmlize(zText, i);
        cgi_printf("%s", z);
        free(z);
        z = htmlize(&zText[i], j);
        cgi_printf("<a href=\"%s\">%s</a>", z, z);
        free(z);
        zText += i+j;
        i = 0;
        wordStart = 0;
        continue;
      }
      if( c=='#' && g.okRead && (j = ndigit(&zText[i+1]))>0 &&
                (v = atoi(&zText[i+1]))>0 && v<=nTicket ){
        z = htmlize(zText, i);
        cgi_printf("%s", z);
        free(z);
        cgi_printf("<a href=\"tktview?tn=%d\">#%d</a>", v, v);
        zText += i+j+1;
        i = 0;
        continue;
      }
      if( c=='[' && g.okCheckout && (j = ndigit(&zText[i+1]))>0 &&
                (v = atoi(&zText[i+1]))>0 && v<=nCommit && zText[i+j+1]==']' ){
        z = htmlize(zText, i);
        cgi_printf("%s", z);
        free(z);
        cgi_printf("<a href=\"chngview?cn=%d\">[%d]</a>", v, v);
        zText += i+j+2;
        i  = 0;
        continue;
      }
    }
    wordStart = isspace(c);
    if( c=='\n' ){
      if( (j = nspace(&zText[i]))>=2 ){
        while( j>0 && (zText[i+j-1]==' ' || zText[i+j-1]=='\t') ){ j--; }
        z = htmlize(zText, i+1);
        cgi_printf("%s", z);
        free(z); 
        if( inPRE ) cgi_printf("</pre>\n");
        inPRE = 0;
        zText += i+j;
        i = -1;
        c = zText[0];
        if( c=='\t' || (c==' ' && (zText[1]==' ' || zText[1]=='\t')) ){
          if( inOL ){ cgi_printf("</ol>\n"); inOL=0; }
          if( inUL ){ cgi_printf("</ul>\n"); inUL=0; }
          cgi_printf("<pre>");
          inPRE = 1;
        }else{
          if( inOL && !is_enum(zText)  ){ cgi_printf("</ol>\n"); inOL=0; }
          if( inUL && !is_bullet(zText) ){ cgi_printf("</ul>\n"); inUL=0; }
          else        cgi_printf("<p>\n");
        }
      }
      if( !inPRE && (zText[i+1]=='*' || isdigit(zText[i+1])) 
                && zText[i+2]==':' && (zText[i+3]==' ' || zText[i+3]=='\t')){
        z = htmlize(zText, i+1);
        cgi_printf("%s", z);
        free(z);
        if( zText[i+1]=='*' ){
          if( inOL ) cgi_printf("</ol>\n");
          if( !inUL ) cgi_printf("<ul compact>\n");
          inUL = 1;
          inOL = 0;
        }else{
          if( inUL ) cgi_printf("</ul>\n");
          if( !inOL ) cgi_printf("<ol compact>\n");
          inUL = 0;
          inOL = 1;
        }
        zText += i+3;
        i = 0;
        cgi_printf("<li>");
        wordStart = 1;
      }
    }
    i++;
  }
  cgi_printf("%h", zText);
  if( inOL ) cgi_printf("</ol>\n");
  if( inUL ) cgi_printf("</ul>\n");
  if( inPRE ) cgi_printf("</pre>\n");
}

/*
** Output text while inserting hyperlinks to ticket and checkin reports.
** Within the text, an occurance of "#NNN" (where N is a digit) results
** in a hyperlink to the page that shows that ticket.  Any occurance of
** [NNN] gives a hyperlink to check-in number NNN.
**
** (Added later:)  Also format the text as HTML.  Insert <p> in place
** of blank lines.  Insert <pre>..</pre> around paragraphs that are
** indented by two or more spaces.  Make lines that begin with "*:"
** or "1:" into <ul> or <ol> list elements.
**
** (Later:)  The formatting is now extended to include all of the
** Wiki formatting options.
*/
void output_formatted(const char *zText){
  output_wiki(zText,"");
}

/*
** Append HTML text to the output that describes the formatting
** conventions implemented by the output_formatted() function
** above.
*/
void append_formatting_hints(void){
  @ <ul>
  @ <li><p>
  @ Blank lines result in a paragraph break.
  @ </p></li>
  @
  @ <li><p>
  @ If a paragraph is indented by a tab or by two or more spaces,
  @ it is displayed verbatim -- in a constant-width font with all
  @ spacing and line breaks preserved.
  @ </p></li>
  @
  @ <li><p>
  @ The characters "<tt>*:</tt>"  (asterisk-colon) at the very beginning
  @ of a line create a bullet in a bullet-list.
  @ The characters "<tt>1:</tt>"  (one-colon) at the very beginning
  @ of a line create a enumeration item in an enumerated list.
  @ </p></li>
  @
  @ <li><p>
  @ Words surrounded by underscores (ex: "<tt>_bold_text_</tt>") are
  @ rendered in a bold font.
  @ </p></li>
  @
  @ <li><p>
  @ Text like "<tt>#123</tt>" or "<tt>[456]</tt>" is replaced by hyperlinks
  @ to ticket #123 and check-in [456], respectively.  Text that looks like
  @ URL or a wiki page name also becomes a hyperlink.
  @ </p></li>
  @
  @ <li><p>
  @ Additional formatting information is available in the Wiki at
  @ <a href="wiki?p=FormattingWikiPages">FormattingWikiPages</a>.
  @ </p></li>
  @ </ul>
}
